`timescale 1ns/1ps
`default_nettype none

module tb();

   wire A_P, A_N, B_P, B_N;
   reg  CLK100, CLK156;

   wire RESET_PB, PMA_INIT, CH_UP, LANE_UP;
   wire TX_TVALID, TX_TREADY, RX_TVALID;
   wire [63:0] TX_TDATA, RX_TDATA;
   
   wire        CLK, RST;
   
   aurora_64b66b_0 uut_aurora
     ( .txp(A_P), .txn(A_N),                       // O
       .rxp(B_P), .rxn(B_N),                       // I

       .init_clk          ( CLK100),               // I
       .gt_refclk1_p      ( CLK156),               // I
       .gt_refclk1_n      (~CLK156),               // I

       .reset_pb          (RESET_PB),              // I
       .pma_init          (PMA_INIT),              // I

       .channel_up        (CH_UP),                 // O
       .lane_up           (LANE_UP),               // O

       .s_axi_tx_tdata    (TX_TDATA),              // I [0:63]
       .s_axi_tx_tvalid   (TX_TVALID),             // I
       .s_axi_tx_tready   (TX_TREADY),             // O
       .m_axi_rx_tdata    (RX_TDATA),              // O [0:63]
       .m_axi_rx_tvalid   (RX_TVALID),             // O

       .user_clk_out      (CLK),                   // O
       .sys_reset_out     (RST),                   // O

       // Normal operation + DRP AXI Lite disabled
       .loopback          (3'b000),                // I [2:0]
       .gt_rxcdrovrden_in (0),                     // I
       .power_down        (0),                     // I

       .s_axi_bready      (0),                     // I
       .s_axi_awvalid     (0),                     // I
       .s_axi_wvalid      (0),                     // I
       .s_axi_arvalid     (0),                     // I
       .s_axi_rready      (0)                      // I
       );

   wire        TXUSERCLK, RXUSERCLK, TX_RDY, RX_RDY, RX_BITSLIP;
   wire [63:0] RXS, TXS, RXDATA, TXDATA;
   wire [5:0]  RXHDRi;
   reg [5:0]   TXHDRi;
   
   gtwizard_ultrascale_0 uut_gtwiz
     ( .gthrxp_in (A_P), .gthrxn_in (A_N),         // I
       .gthtxp_out(B_P), .gthtxn_out(B_N),         // O

       .gtwiz_reset_clk_freerun_in  (CLK100),      // I
       .gtrefclk00_in               (CLK156),      // I
       .gtwiz_userclk_tx_usrclk2_out(TXUSERCLK),   // O
       .gtwiz_userclk_rx_usrclk2_out(RXUSERCLK),   // O

       .gtwiz_reset_all_in          (RESET_PB),    // I
       .gtwiz_reset_tx_done_out     (TX_RDY),      // O
       .gtwiz_reset_rx_done_out     (RX_RDY),      // O
       .gtwiz_userdata_tx_in        (TXS),         // I [63:0]
       .gtwiz_userdata_rx_out       (RXS),         // O [63:0]
       .rxgearboxslip_in            (RX_BITSLIP),  // I
       .txheader_in                 (TXHDRi),      // I [5:0]
       .rxheader_out                (RXHDRi),      // O [5:0]

       .gtwiz_userclk_rx_reset_in   (1'b0),        // I
       .gtwiz_userclk_tx_reset_in   (1'b0),        // I
       .gtwiz_reset_tx_pll_and_datapath_in(1'b0),  // I
       .gtwiz_reset_tx_datapath_in        (1'b0),  // I
       .gtwiz_reset_rx_pll_and_datapath_in(1'b0),  // I
       .gtwiz_reset_rx_datapath_in        (1'b0)   // I
       );

   teng_sc   sc ( .CLK(TXUSERCLK), .RST(~TX_RDY), .D(TXDATA), .S(TXS) );
   teng_desc ds ( .CLK(RXUSERCLK), .RST(~RX_RDY), .S(RXS), .D(RXDATA) );

   reg [1:0] RXHDR;
   wire [1:0] TXHDR;
   always @ (TXUSERCLK) RXHDR <= RXHDRi[1:0];
   always @ (RXUSERCLK) TXHDRi <= TXHDR;

   wire [3:0]  RXSTAT;

   tx_engine txe ( .CLK(TXUSERCLK), .RST(~TX_RDY), .RXSTAT(RXSTAT),
                   .D(TXDATA), .H(TXHDR) );

   rx_engine rxe ( .CLK(RXUSERCLK), .RST(~RX_RDY), .RXSTAT(RXSTAT),
                   .D(RXDATA), .H(RXHDR), .BITSLIP(RX_BITSLIP) );
   
   initial begin
      $dumpfile("a.vcd");
      $dumpvars(1);

      CLK100 <= 1;  CLK156 <= 1; end
   
   always # (5)   CLK100 <= ~CLK100;
   always # (3.2) CLK156 <= ~CLK156;

   ku_aurora_boot boot ( .CLK100(CLK100),     .DCM_LOCKED(1),
                         .PMA_INIT(PMA_INIT), .RESET_PB(RESET_PB) );

   tx_gen tg ( .CLK(CLK), .RST(RST), 
               .TREADY(TX_TREADY & rxe.RXSTAT[3]), 
               .TVALID(TX_TVALID), .TDATA(TX_TDATA) );
   
endmodule // tb

module ku_aurora_boot
  ( input wire CLK100,
    input wire DCM_LOCKED,
    output reg PMA_INIT, RESET_PB
    );
   
   reg [7:0]        PMA_INIT_CNT = 0;

   initial PMA_INIT <= 1;
   initial RESET_PB <= 1;
   
   always @ (posedge CLK100) begin
      if (~DCM_LOCKED) PMA_INIT_CNT <= 0;
      else PMA_INIT_CNT <= PMA_INIT_CNT + ((~&PMA_INIT_CNT) ? 1 : 0);
      
      case (PMA_INIT_CNT)
        100: PMA_INIT <= 0;
        200: RESET_PB <= 0;
      endcase
   end 
endmodule

module tx_gen
  ( input wire CLK, RST,
    input wire        TREADY,
    output reg        TVALID,
    output reg [63:0] TDATA );

   parameter PayloadLen=100;
   
   always @ (posedge CLK) begin
      if (RST) begin
         TDATA <=0; TVALID <= 0;
      end else begin
         if (TREADY) begin
            TVALID <= (TDATA==0) ? 1 : (TDATA==PayloadLen) ? 0 : TVALID;
            if (TDATA!=PayloadLen) TDATA <= TDATA+1;
         end
      end
   end
endmodule

module teng_sc
  ( input wire CLK, RST,
    input wire [65:128]  D,
    output wire [ 0:63] S );

   wire [1:128]         Sc; // S combinational
   reg [0:63]           Sr = 0; // S register

   assign Sc[ 1: 64] = S[ 0:63];
   assign Sc[65:128] = D[65:128] ^ Sc[26:89] ^ Sc[7:70];

   always @ (posedge CLK) begin
      if (RST) Sr <= 0;
      else     Sr <= Sc[65:128];
   end

   assign S = Sr;

endmodule // teng_sc

module teng_desc
  ( input wire CLK, RST,
    input wire [65:128]  S,
    output reg [0:63] D );

   // S[16:31] = scrambled input
   // S[ 1:15] = previous scrambled input = Sr[1:15]

   reg [1:64]           Sr;
   wire [1:128]         Sc;

   always @ (posedge CLK)
     if (RST) Sr <= 0;
     else     Sr[1:64] <= S[65:128];

   assign Sc[ 1: 64] = Sr[1:64];
   assign Sc[65:128] = S[65:128];

   always @ (posedge CLK)
     D[0:63] <= Sc[26:89] ^ Sc[7:70] ^ Sc[65:128];

endmodule // teng_desc

module tx_engine 
  ( input wire CLK, RST,
    input wire [3:0]   RXSTAT,
    output wire [63:0] D,
    output wire [1:0]  H );

   // CB timer for Not ready - Wait for remote : Once in 16 clk
   reg [3:0]           CB_TIMER;
   wire                SEND_CB = &CB_TIMER;
   always @ (posedge CLK) CB_TIMER <= RST ? 0 : CB_TIMER+1;

   // CC timer for Channel up : 7 in 512 clk (too frequent but OK)
   reg [8:0]           CC_TIMER;
   reg                 SEND_CC;
   always @ (posedge CLK) begin
      CC_TIMER <= RST ? 0 : CC_TIMER+1;
      SEND_CC  <= (CC_TIMER==1) ? 1 : (CC_TIMER==7) ? 0 : SEND_CC; end

   // Rx state synchronizer
   reg [3:0] RXSTATi, RXSTAT_TX;
   always @ (posedge CLK) begin
      RXSTAT_TX <= RXSTATi;  RXSTATi <= RXSTAT; end 

   assign D = RXSTAT_TX[0] ? {  16'h7830, 48'h0} :   // Not ready: NR+SA
              RXSTAT_TX[1] ? ( SEND_CB ?             // Channel bonding:
                               {16'h7840, 48'h0} :   //   CB
                               {16'h7830, 48'h0} ) : //   NR+SA
              RXSTAT_TX[2] ? ( SEND_CB ?             // Wait for remote
                               {16'h7840, 48'h0} :   //   CB
                               {16'h7810, 48'h0} ) : //   SA
              RXSTAT_TX[3] ? ( SEND_CC ?             // Channel up
                               {16'h7880, 48'h0} :   //   CC
                               {16'h7810, 48'h0} ) : //   SA
              0;                                     // Never come here

   assign H = 2'b10;  // always control

endmodule // tx_engine

module rx_engine 
  ( input wire CLK, RST,
    output reg [3:0] RXSTAT,
    input wire [63:0] D,
    input wire [1:0]  H,
    output wire       BITSLIP );

   initial RXSTAT <= 1;

   wire        RX_IS_IDLE = (H[1:0] == 2'b10 & D[63:56]==8'h78 );
   wire        RX_IS_CC = RX_IS_IDLE & D[55]; // Clock comp
   wire        RX_IS_CB = RX_IS_IDLE & D[54]; // Ch bond
   wire        RX_IS_NR = RX_IS_IDLE & D[53]; // Not ready
   wire        RX_IS_SA = RX_IS_IDLE & D[52]; // Strict alignment
   
   reg [6:0]   RXSLIP_TIMER;
   reg [9:0]   CNT;
   wire [3:0]  RXSTAT_NEXT = {RXSTAT[2:0], RXSTAT[3]};
   
   always @ (posedge CLK) begin
      if (RST) begin
         RXSTAT <= 4'b0001;
         RXSLIP_TIMER <= 0;
         CNT  <= 0;
      end else begin
         case (RXSTAT)
           'b0001: begin // Not ready
              CNT  <= (RX_IS_IDLE) ? CNT + 1 : 0;
              RXSLIP_TIMER <= RXSLIP_TIMER + 1;
              if (CNT==16) begin
                 CNT <= 0;
                 RXSTAT <= RXSTAT_NEXT;  end end

           'b0010: begin // Channel bonding: but no CB check for single-lane
              if (~(RX_IS_CC | RX_IS_CB | RX_IS_NR) | CNT == 50 ) begin
                 RXSTAT <= RXSTAT_NEXT;
                 CNT <= 0;
              end else begin
                 if (RX_IS_CB) CNT <= CNT+1; end end

           'b0100: begin // Wait for remote
              if (CNT==200) begin
                 RXSTAT <= RXSTAT_NEXT;
                 CNT <= 0;
              end else begin
                if (~RX_IS_NR & RX_IS_IDLE) CNT <= CNT+1; 
                else CNT <= 0; end end

           'b1000:  // Channel up
             RXSTAT <= RXSTAT;
              
           default: RXSTAT <= 'b0001;
         endcase
      end
   end
   
   assign BITSLIP = &RXSLIP_TIMER & RXSTAT[0];

   // Rx data
   wire        RX_IS_DATA = (H[1:0] == 2'b01) & RXSTAT[3];

   wire [63:0]        RX_DATA;
   genvar             i;
   for (i=0; i<8; i=i+1) begin : rx_data_gen
      assign RX_DATA[i*8+7 : i*8] = D[63-i*8 : 56-i*8];
   end

endmodule // rx_engine

`default_nettype wire


/*
// Aurora 64B/66B unconnected
 
        // Slave AXI disabled
       .s_axi_wstrb       (),                      // I [3:0]
       .s_axi_awaddr      (),                      // I [31:0]
       .s_axi_rresp       (),                      // O [1:0]
       .s_axi_bresp       (),                      // O [1:0]
       .s_axi_wdata       (),                      // I [31:0]
       .s_axi_araddr      (),                      // I [31:0]
       .s_axi_rdata       (),                      // O [31:0]
       .s_axi_awready     (),                      // O
       .s_axi_arready     (),                      // O
       .s_axi_rvalid      (),                      // O
       .s_axi_wready      (),                      // O
       .s_axi_bvalid      (),                      // O
       
       // Slave signals
       .tx_out_clk        (),                      // O
       .gt_pll_lock       (),                      // O
       .mmcm_not_locked_out(),                     // O
       .link_reset_out    (),                      // O
       .sync_clk_out      (),                      // O
       .gt_qpllclk_quad1_out(),                    // O
       .gt_qpllrefclk_quad1_out(),                 // O
       .gt_qpllrefclklost_quad1_out(),             // O
       .gt_qplllock_quad1_out(),                   // O
       .gt_reset_out      (),                      // O
       .gt_refclk1_out    (),                      // O

       // Less used status signals
       .hard_err          (),                      // O
       .soft_err          (),                      // O
       .gt_powergood      ()                       // O  [0:0]
 
 // GTH wiz unconnected
         
       .txsequence_in(),                           // I [6:0]
       .rxdatavalid_out(),                         // O [1:0]
       .rxheadervalid_out(),                       // O [1:0]
       .rxprgdivresetdone_out(),                   // O

       .qpll0outclk_out            (),             // O
       .qpll0outrefclk_out         (),             // O
       .gtwiz_userclk_rx_active_out(),             // O
       .gtwiz_userclk_tx_srcclk_out(),             // O
       .gtwiz_userclk_tx_usrclk_out(),             // O
       .gtwiz_userclk_tx_active_out(),             // O
       .gtwiz_userclk_rx_srcclk_out(),             // O
       .gtwiz_userclk_rx_usrclk_out(),             // O
       .gtwiz_reset_rx_cdr_stable_out(),           // O
       .gtpowergood_out(),                         // O
       .rxstartofseq_out(),                        // O [1:0]
       .rxpmaresetdone_out(),                      // O
       .txpmaresetdone_out(),                      // O
       .txprgdivresetdone_out()                    // O
       );
 */
